#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys, math, scipy.constants as sp
from collections import Counter
from rdkit import Chem
from rdkit.Chem import rdMolDescriptors

# r.h.s of Frohlich equation (dimensionless) = alpha*[Debye**2]/(Kelvin*cc/mol)
alpha = sp.N_A*(3.33564*10**(-30))**2/(9*sp.epsilon_0*sp.k*10**(-6))

# vi, ri <= parameters needed to estimate molar volume Vm and refrative index nD
params = (
	('B30', 7.61, 2.91),
	('Br10', 27.52, 8.44),
	('C10', 14.68, 2.32),
	('C20', 9.89, 4.01),
	('C21', 22.77, 4.58),
	('C30', -0.00, 3.15),
	('C30a', -0.00, 3.48),
	('C31', 13.23, 4.65),
	('C31a', 13.23, 4.46),
	('C32', 27.17, 5.38),
	('C40', -8.40, 2.60),
	('C41', 4.46, 3.48),
	('C42', 16.57, 4.60),
	('C43', 29.58, 5.74),
	('Cl10', 24.74, 5.87),
	('F10', 17.63, 1.06),
	('Ge40', 9.36, 8.54),
	('I10', 35.64, 13.75),
	('N10', 14.09, 1.55),
	('N20', 7.42, 2.60),
	('N20a', 7.42, 2.44),
	('N21', 18.14, 3.55),
	('N30', -3.08, 2.89),
	('N30a', -3.08, 2.85),
	('N31', 7.74, 3.69),
	('N31a', 7.74, 3.72),
	('N32', 17.81, 4.60),
	('N43', 5.36, 8.02),
	('O10', 14.89, 1.84),
	('O20', 6.25, 1.55),
	('O20a', 6.25, 0.71),
	('O21', 11.78, 2.51),
	('P30', 10.42, 7.41),
	('P40', -1.94, 4.98),
	('P41', 10.06, 5.41),
	('R5', 9.41, None),
	('R6', 6.89, None),
	('R<5', 10.89, None),
	('R>6', 3.75, None),
	('S10', 25.92, 10.60),
	('S20', 14.90, 8.22),
	('S20a', 14.90, 7.05),
	('S21', 26.14, 8.71),
	('S30', 5.58, 7.28),
	('S40', -3.74, 5.12),
	('Se20', 19.00, 11.53),
	('Se20a', 19.00, 9.43),
	('Si40', 9.28, 6.38),
	('Si41', 23.35, 7.85),
	('Sn40', 14.47, 14.90),
	('Ti40', 6.09, 14.66),
	('aromat', 1.82, None))
vi = dict((p[0], p[1]) for p in params)
ri = dict((p[0], p[2]) for p in params)

# A_dic, B_dic <= nominal (i.e. neglecting interactions between fragments)
# parameters associated with polar fragments to estimate dielectric constant
Nominal = (
	('c1cnccn1',0,0.28),
	('COB(OC)OC',0,0.52),
	('C[Si](C)(C)C',0,0.12),
	('CO[Si](C)(C)C',0,1.31),
	('CO[Si](C)(C)OC',0,1.78),
	('CO[Si](C)(OC)OC',0,2.51),
	('CO[Si](OC)(OC)OC',0,2.74),
	('Fc1ccccc1',0,2.08),
	('Clc1ccccc1',0,2.19),
	('Brc1ccccc1',0,1.91),
	('CN',0,1.90),
	('Nc1ccccc1',0,2.26),
	('CNC',0,1.10),
	('CNc1ccccc1',0,2.69),
	('CN(C)C',0,0.49),
	('CN(C)c1ccccc1',0,1.99),
	('COC',0,1.74),
	('COC_ring',0,3.13),
	('COc1ccccc1',0,1.43),
	('CS',0,2.33),
	('CSC',0,2.55),
	('CSc1ccccc1',0,1.92),
	('COC=O',0,3.51),
	('COC(C)=O',0,2.95),
	('CC(=O)Oc1ccccc1',0,2.35),
	('COC(=O)c1ccccc1',0,3.33),
	('COC(=O)/C=C/c1ccccc1',0,3.71),
	('C=CC(=O)OC',0,3.25),
	('COC(=O)Cl',0,5.82),
	('COC(=O)CCl',0,7.14),
	('COC(=O)C(Cl)(Cl)Cl',0,5.59),
	('COC(=O)CBr',0,5.94),
	('COC(=O)c1ccccc1C(=O)OC',0,6.61),
	('CON=O',0,5.61),
	('c1ccncc1',0,4.84),
	('c1ccc2ncccc2c1',0,4.10),
	('COP(C)(=O)OC',0,8.64),
	('COP(=O)(OC)OC',0,12.74),
	('COC(N)=O',0,7.68),
	('CN(C)C(=O)N(C)C',0,15.31),
	('CN=C=O',0,8.04),
	('C[N+](=O)[O-]',0,12.68),
	('O=[N+]([O-])c1ccccc1',0,15.14),
	('CO[N+](=O)[O-]',0,8.90),
	('CC(=O)OC(C)=O',0,12.28),
	('CC=O',0,7.09),
	('CC(C)=O',0,9.44),
	('CC(C)=O_ring',0,8.27),
	('CC#N',0,12.13),
	('CN(C)C=O',0,19.04),
	('CC(=O)N(C)C',0,22.02),
	('CS(C)=O',0,24.50),
	('CS(C)(=O)=O',0,29.59),
	('CS(=O)(=O)c1ccccc1',0,30.15),
	('CS(C)(=O)=O_ring',0,22.57),
	('[O-][n+]1ccccc1',0,17.91),
	('COC(C)=O_ring',0,20.28),
	('CP(C)(C)=O',0,37.49),
	('CP(C)(C)=S',0,42.81),
	('CNC(=O)OC_ring',0,31.24),
	('COC(=O)N(C)C_ring',0,41.89),
	('COC(=O)OC',0,0.89),
	('COC(=O)OC_ring',0,36.41),
	('C[n+]1cc([O-])on1',0,50.82),
	('CC(=O)O',0,0.78),
	('Oc1ccccc1',0,3.49),
	('COC(=O)c1ccccc1O',0,4.43),
	('CF',-0.64,3.46),
	('CCl',-1.02,5.83),
	('CBr',-1.19,5.97),
	('CI',-0.94,4.88),
	('CNC=O',16.90,22.18),
	('CNC(C)=O',8.12,69.27),
	('CO',0,9.19))
A_dic = dict((p[0], p[1]) for p in Nominal)
B_dic = dict((p[0], p[2]) for p in Nominal)

# T_dic <= parameters needed to estimate interactions between polar fragments
T_dic = {
	'CN':0.41,
	'COC':-0.10,
	'CS':-0.41,
	'COC(C)=O':-0.22,
	'CO[N+](=O)[O-]':-0.05,
	'CC(C)=O':-0.12,
	'CC#N':0.03,
	'CC(=O)O':1.59,
	'CCl':-0.50,
	'CBr':-0.73,
	'CO':-0.35}

def get_atom_code(atom):
	""" return an atom code consistent with the keys of the 'vi' dictionary """
	# be careful to take into account nD = number of deuterium atoms !
	nD = len([x for x in atom.GetNeighbors() if x.GetMass()>2 and x.GetSymbol()=='H'])
	code = atom.GetSymbol() + str(atom.GetTotalDegree()) + str(atom.GetTotalNumHs()+nD)
	code += 'a'*int(atom.GetIsAromatic())
	return code

def get_ring_descriptors(mol, maxi=6, mini=5):
	""" return dict of ring descriptors for molecule provided as input """
	dic = Counter()
	ri = mol.GetRingInfo()
	for ring in ri.AtomRings():
		size = len(ring)
		if size > maxi:
			label = 'R>'+str(maxi)
		elif size < mini:
			label = 'R<'+str(mini)
		else:
			label = 'R%u' %len(ring)
		dic[label] += 1
		# specific contribution of aromatic rings
		atoms = [mol.GetAtomWithIdx(i) for i in ring]
		if all(at.GetIsAromatic() for at in atoms):
			dic['aromat'] += 1
	return dic

def GetPolarFragments(mol, tau=1.7):
	"""Renvoie la liste des fragments POLAIRES de la molécule sous la forme d'une liste,
	chaque fragment étant défini sous la forme : (Nom,Correction,Nombre d'hydrogène)"""
	# split the molecule on bonds => Fragments, Indices
	patt = Chem.MolFromSmarts('[CX4][$([CX4]),$([c])]')
	bonds = mol.GetSubstructMatches(patt)
	em = Chem.EditableMol(mol)
	nAts = mol.GetNumAtoms()
	for a,b in bonds:
		em.RemoveBond(a,b)
		em.AddAtom(Chem.Atom(1))
		em.AddBond(a,nAts,Chem.BondType.SINGLE)
		em.AddAtom(Chem.Atom(1))
		em.AddBond(b,nAts+1,Chem.BondType.SINGLE)
		nAts+=2
	p = em.GetMol()
	p = Chem.RemoveHs(p)
	Chem.SanitizeMol(p)
	Fragments = Chem.GetMolFrags(p,asMols=True)
	Indices = Chem.GetMolFrags(p,asMols=False)
	# prepare list of polar fragments in mol => Fragments_polaires
	Fragments_polaires = [] # prepare list of all polar fragments
	# prepare heteroatoms as minimal polar fragments
	patt = Chem.MolFromSmarts('[N,n,O,o,P,S,s,Si,B,Sn,F,Cl,Br,I]')
	for frag,idx in zip(Fragments,Indices):
		# require that polar fragment contain a heteroatom
		if len(frag.GetSubstructMatches(patt)) > 0:  # polar fragments must contain an heteroatom
			nom = Chem.MolToSmiles(frag)             # used as fragment name
			# now add suffix '_ring' to fragment name if needed
			Ring = 0
			for atom in idx:
				if mol.GetAtomWithIdx(atom).IsInRing() and not mol.GetAtomWithIdx(atom).GetIsAromatic():
					Ring += 1
			if Ring > 2 :
				nom = nom + '_ring'
			# set N_h <= number of hydrogens affecting the fragment contrib. to gmu2
			if nom == 'CF' or nom == 'CCl' or nom == 'CBr' or nom == 'CI':
				for atom in idx:
					if p.GetAtomWithIdx(atom).GetSymbol() == 'C':
						N_h = mol.GetAtomWithIdx(atom).GetTotalNumHs() #Pas de changement des numéros d'atomes avec le p
			elif nom == 'CNC=O' or nom == 'CNC(C)=O':
				for atom in idx:
					if p.GetAtomWithIdx(atom).GetSymbol() == 'C' :
						for neighbor in p.GetAtomWithIdx(atom).GetNeighbors():
							if neighbor.GetSymbol() == 'N' and p.GetAtomWithIdx(atom).GetTotalDegree() == 4:
								N_h = mol.GetAtomWithIdx(atom).GetTotalNumHs()
			else:
				N_h = 0
			# delta <= correction due to inter-fragment interactions
			delta = 0
			for frag_voisin,idx_voisin in zip(Fragments,Indices):
				# skip hydrocarbons as neighboring fragments
				if idx != idx_voisin and len(frag_voisin.GetSubstructMatches(patt)) > 0:
					Distance = 9.E9    # defaut interfragment distance very large (yielding zero interaction)
					for atom in idx :
						if p.GetAtomWithIdx(atom).GetSymbol() != '*':
							for atom_voisin in idx_voisin:
								if p.GetAtomWithIdx(atom_voisin).GetSymbol() != '*':
									Distance = min(len(Chem.rdmolops.GetShortestPath(mol,atom,atom_voisin)),Distance)
					delta += math.exp(-(Distance-2)/tau) # tij - 1 = (Nombre d'atomes - 1) - 1  as last fragment atom is counted in Distance
			Fragments_polaires.append((nom,delta,N_h))
	return Fragments_polaires

def get_calc_gmu2(mol):
	Fragments_polaires = GetPolarFragments(mol)
	frags = Counter(nom for nom,delta,N_h in Fragments_polaires)
	missing_frags = set(frags) - set(B_dic)
	if missing_frags:
		return 'Missing fragments: ' + ' '.join(sorted(missing_frags))
	Total = len(Fragments_polaires)
	if Total > 1:
		missing_frags = set(frags) - set(T_dic)
		if missing_frags:
			return 'Missing fragments: ' + ' '.join(sorted(missing_frags))
	gu2 = 0
	for nom,delta,N_h in Fragments_polaires :
		gu2 += A_dic[nom]*N_h + B_dic[nom]
		if Total > 1 and nom in T_dic:
			gu2 *= 1 + T_dic[nom]*delta/(Total-1)
	return gu2

def get_properties(mol):
	""" return molar_volume, liq_density, refractive_index for mol """
	atoms = Counter(get_atom_code(atom) for atom in mol.GetAtoms())
	if set(atoms) - set(vi): return None, None, None
	mw = rdMolDescriptors.CalcExactMolWt(mol)
	rings = get_ring_descriptors(mol)
	Vm = sum([atoms[k]*vi[k] for k in atoms])   # atoms contribution
	Vm += sum([rings[k]*vi[k] for k in rings])  # rings contribution
	density = mw/Vm
	if set(atoms) - set(ri): return Vm, density, None
	RD = sum([atoms[k]*ri[k] for k in atoms])   # molar refraction
	nD = ((Vm+2*RD)/(Vm-RD))**0.5
	return Vm, density, nD

def get_calc_gmu2(mol):
	""" return g*mu^2 as obtained from a sum of polar fragment contributions """
	Fragments_polaires = GetPolarFragments(mol)
	frags = Counter(nom for nom,delta,N_h in Fragments_polaires)
	missing_frags = set(frags) - set(B_dic)
	if missing_frags:
		return '=> missing gmu2 contribution for: %s' %' '.join(sorted(missing_frags))
	Total = len(Fragments_polaires)
	if Total > 1:
		missing_frags = set(frags) - set(T_dic)
		if missing_frags:
			return '=> missing interaction parameter for: %s' %' '.join(str(w) for w in missing_frags)
	gu2 = 0
	for nom,delta,N_h in Fragments_polaires :
		if Total == 1:
			gu2 += A_dic[nom]*N_h + B_dic[nom]
		elif Total > 1:
			try:
				gu2 += (A_dic[nom]*N_h + B_dic[nom])*(1+T_dic[nom]*delta/(Total-1))
			except:
				gu2 += A_dic[nom]*N_h + B_dic[nom]
	return gu2

def get_dielectric_constant(mol, TK, Vm, nD):
	gu2 = get_calc_gmu2(mol)
	if type(gu2) is str: return gu2
	x = alpha * gu2 / (TK * Vm)
	a = nD**2
	Delta2 = 8*a**2 + (a + x*(a+2)**2)**2
	DC = 0.25*(a + x*(a+2)**2 + math.sqrt(Delta2))
	return DC

def read_input_line(line):
	""" return molecule and temperature(K) from line """
	data = line.split()
	if not data: return None, None
	mol = Chem.MolFromSmiles(data[0])
	if not mol: return None, None
	if len(data) == 1: return mol, 298.15
	try:
		celsius = float(data[1])
	except ValueError:
		celsius = 25.
	return mol, sp.convert_temperature(celsius, 'Celsius', 'Kelvin')

def get_output_line(density, Vm, nD, dc):
	if type(density) is float:
		line = '%5.3f %6.1f' %(density, Vm)
	else:
		line = '%s %s' %(str(density), str(Vm))
	if type(nD) is float:
		line += ' %6.3f' %nD
	else:
		line += ' %s' %str(nD)
	try:
		line += ' %8.4f' %dc
	except ValueError:
		line += ' %s' %str(dc)
	return line

msg = """Every line in input file provided as argument must contain:
* a SMILES string encoding the molecule
* the requested temperature in °C (optional, default is 25°C)
An output line is returned for every input line, with:
* 'No molecule' as message if the molecule could not be read
* otherwise, blank-separated data, namely:
<density(cc/mol)> <molar vol.(cc/mol)> <refractive index> <dielectric constant>
"""

if __name__ == '__main__':
	if len(sys.argv) < 2:
		print(msg)
	else:
		lines = open(sys.argv[1]).readlines()
		if lines[-1] == '\n': lines = lines[:-1]
		for line in lines:
			mol, TK = read_input_line(line)
			if not mol:
				print('No molecule')
			else:
				Vm, density, nD = get_properties(mol)
				if None in (Vm, nD):
					dc = None
				else:
					dc = get_dielectric_constant(mol, TK, Vm, nD)
				print(get_output_line(density, Vm, nD, dc))
